import numpy as np
import casadi as ca
import spatial_casadi as sc
from scipy.spatial.transform import Rotation as R
from grasp_planning.solver.robot_model import RobotKinematicModel
from grasp_planning.cost.costs import SquaredAccCost
from grasp_planning.constraints.constraints import GraspPosConstraint, GraspRotConstraint, ManipulationConstraint

class GOMP():
    def __init__(self, n_waypoints, urdf, theta=np.pi/4, root_link='world', end_link='tool_frame') -> None:
        # Init variables
        self.T_W_Grasp = None
        self.T_W_Obj = None
        self.T_Obj_Grasp = np.eye(4, dtype=float)
        
        # Kinematics
        self._robot_model = RobotKinematicModel(urdf, root_link, end_link)
        self.n_dofs = self._robot_model.n_dofs
        self.manipulation_frame_dim = 6
        self.x_dim = self.n_dofs + self.manipulation_frame_dim
        self.n_waypoints = n_waypoints
        self.theta = theta
        
        # Optimization
        self.x = ca.SX.sym("x", self.x_dim, self.n_waypoints)
        self.x_init = None
        
        self._objective = None
        self.l_joint_limits, self.u_joint_limits = None, None
        self.g_list = []




    def update_grasp_DOF(self, theta: float, degrees=True) -> None:
        """
        Object has to have z-axis aligned with the world frame
        """
        if degrees:
            R_Obj_Grasp = R.from_euler('xyz', [180, 0, theta], degrees=True).as_matrix()
            self.theta = np.deg2rad(theta)
        else:
            R_Obj_Grasp = R.from_euler('xyz', [np.pi, 0, theta], degrees=False).as_matrix()
            self.theta = theta
        self.T_Obj_Grasp[:3,:3] = R_Obj_Grasp
        self.T_W_Grasp = self.T_W_Obj @ self.T_Obj_Grasp

    def update_object_pose(self, T_W_Obj: np.array) -> None:
        """
        Update object and grasp poses
        """
        self.T_W_Obj = T_W_Obj

    def _grasp2rpy(self):
        """
        Convert T_W_Grasp into [x,y,z,r,p,y]
        """
        R_W_Grasp_rpy = sc.Rotation.from_matrix(self.T_W_Grasp[:3,:3]).as_euler("xyz") #convert rotation part of FK into rpy 
        self.T_W_Grasp_rpy = np.vstack((self.T_W_Grasp[:3,3][:,None], R_W_Grasp_rpy))
        return self.T_W_Grasp_rpy


    def set_init_guess(self, q):
        self.x_init = np.tile(q, self.n_waypoints)

    def set_boundary_conditions(self, q_start, q_end=None):
        if self.l_joint_limits is None or self.u_joint_limits is None:
            # Reset joint limits
            self.l_joint_limits = np.full((self.x_dim, self.n_waypoints), -100.0)
            self.u_joint_limits = np.full((self.x_dim, self.n_waypoints), 100.0)

            # get joint limits
            joint_limits = self._robot_model.get_joint_pos_limits()
            for t in range(self.n_waypoints):
                self.l_joint_limits[:self.n_dofs, t] = joint_limits[:,0]
                self.u_joint_limits[:self.n_dofs, t] = joint_limits[:,1]
                self.l_joint_limits[-3:, t] = -2*np.pi
                self.u_joint_limits[-3:, t] = 2*np.pi

        
        # init boundary
        self.l_joint_limits[:,0] = q_start
        self.u_joint_limits[:,0] = q_start

        if q_end != None:
            print("Final boundary condition is not implemented.")
      

    def setup_problem(self):
        # assert self.T_Obj_Grasp != None, "Grasp DOF is not set. Set additional degree of freedom around object."
        # assert self.T_W_Obj != None, "Object pose is not set."
        # 1. create decision variable
        x_ca_flatten = self.x.reshape((-1,1))
        
        # assert self.x_init != None, "Initial guess is missing. Set up one using `set_init_guess(q)`."

        # Define objective function
        cost = SquaredAccCost(self.n_waypoints, self.x_dim, manip_frame=True)
        objective = cost.eval_cost(x_ca_flatten)

        # Define constraints
        g = ca.vertcat()
        for g_term in self.g_list:
            gt = g_term.get_constraint()
            g = ca.vertcat(g, gt)

        options = {}
        options["ipopt.acceptable_tol"] = 1e-3
        self.solver = ca.nlpsol('solver', 'ipopt', {'x': x_ca_flatten, 'f': objective, 'g': g}, options)
  
        
    def solve(self):
        result = self.solver(x0=self.x_init,
                             lbg=self.g_lb,
                             ubg=self.g_ub,
                             lbx=self.l_joint_limits.reshape((-1,1), order='F'),
                             ubx=self.u_joint_limits.reshape((-1,1), order='F'))
        return result['x']

    def _add_grasp_pos_constraint(self, waypoint_ID, tolerance=0.0):
        self.g_list.append(ManipulationConstraint(self._robot_model, self.x, waypoint_ID, self.n_dofs))
        self.g_list.append(GraspPosConstraint(self._robot_model, self.x, waypoint_ID, self.n_dofs, tolerance))

    def _add_grasp_rot_constraint(self, waypoint_ID, tolerance=0.1):
        self.g_list.append(GraspRotConstraint(self._robot_model, self.x, waypoint_ID, self.n_dofs, self.theta, tolerance))


    def add_grasp_constraint(self, waypoint_ID, tolerance=0.0):
        self._add_grasp_pos_constraint(waypoint_ID, tolerance)
        self._add_grasp_rot_constraint(waypoint_ID)

    def add_collision_constraint(self, waypoint_ID):
        pass

    def update_constraints_boundaries(self, T_W_Obj):
        self.update_object_pose(T_W_Obj)
        self.update_grasp_DOF(theta=self.theta, degrees=False)
   
        self.g_lb, self.g_ub = ca.vertcat(), ca.vertcat()

        for g_term in self.g_list:
            g_term.update_limits(self.T_W_Grasp)
            g_lb, g_ub = g_term.get_limits()
            self.g_lb = ca.vertcat(self.g_lb, g_lb)
            self.g_ub = ca.vertcat(self.g_ub, g_ub)
        




import numpy as np
import casadi as ca
import spatial_casadi as sc
from grasp_planning.constraints.constraint_template import Constraint

class GraspRotConstraint(Constraint):
    def __init__(self, robot, q_ca, waypoint_ID, n_dofs, theta=0.0, tolerance=0.1) -> None:
        super().__init__() 
        # q = q_ca[:n_dof, waypoint_ID] -> joint space of the robot (decision variable)
        # manip_frame = q_ca[n_dof:, waypoint_ID] -> manipulation frame (decision variable)
        # manip frame_pos - T_W_Grasp_pos = 0
        self._robot = robot
        self.theta = theta
        self.tolerance = tolerance

        self.q_ca = q_ca
        fk_ca_T_rpy = self._robot.compute_fk_rpy_ca(q_ca[:n_dofs, waypoint_ID])
        R_W_EEF_rpy = fk_ca_T_rpy[3:]



        # self._g_x_rot = R_W_EEF_rpy[0]
        # self._g_y_rot = R_W_EEF_rpy[1]
        # self._g_z_rot = R_W_EEF_rpy[2]
        fk_ca_T = self._robot.compute_fk_ca(q_ca[:n_dofs, waypoint_ID])

        # R_G_EEF = R_W_G.T @ fk_ca_T[:3,:3]
        R_W_EEF = fk_ca_T[:3,:3]


    
        self.g = sc.Rotation.from_matrix(R_W_EEF).as_euler("xyz")


        # rpy = sc.Rotation.from_matrix(R_G_EEF).as_euler("xyz")
        # self._g_x_rot = rpy[0]
        # self._g_y_rot = rpy[1]
        # self._g_z_rot = rpy[2]
        # g_y_l = -theta
        # g_y_u = theta
        # g_x_l = -0.1
        # g_x_u = 0.1
        # g_z_l = -0.1
        # g_z_u = 0.1
        # # self.update_limits(T_W_Grasp)
        # self.g_lb = ca.vertcat(g_x_l, g_y_l, g_z_l)
        # self.g_ub = ca.vertcat(g_x_u, g_y_u, g_z_u)
        # self.g = ca.vertcat(self._g_x_rot, self._g_y_rot, self._g_z_rot)
        
        self.g_eval = ca.Function('g_grasp_rot', [self.q_ca], [self.g])
    
    def get_constraint(self):
        return self.g

    def do_eval(self, q):
        return self.g_eval(q)

    def set_limits(self, lb, ub):
        self.g_lb = lb
        self.g_ub = ub

    def get_limits(self):
        return self.g_lb, self.g_ub
    
    def update_limits(self, T_W_Grasp):
        # R_W_G_rpy = sc.Rotation.from_matrix(T_W_Grasp[:3,:3]).as_euler("xyz")
        # # x-axis
        # self._g_x_l = R_W_G_rpy[0] - self.tolerance 
        # self._g_x_u = R_W_G_rpy[0] + self.tolerance
        # # y-axis
        # self._g_y_l = R_W_G_rpy[1] - self.theta 
        # self._g_y_u = R_W_G_rpy[1] + self.theta 
        # # z-axis
        # self._g_z_l = R_W_G_rpy[2] - self.tolerance 
        # self._g_z_u = R_W_G_rpy[2] + self.tolerance 

        # g_lb = ca.vertcat(self._g_x_l, self._g_y_l, self._g_z_l)
        # g_ub = ca.vertcat(self._g_x_u, self._g_y_u, self._g_z_u)
        # self.set_limits(g_lb, g_ub)
        
        R_W_G = T_W_Grasp[:3,:3]
        R_lb = sc.Rotation.from_euler("xyz", [-self.tolerance, -self.theta, -self.tolerance], degrees=False).as_matrix()
        R_ub = sc.Rotation.from_euler("xyz", [self.tolerance, self.theta, self.tolerance], degrees=False).as_matrix()

        R_lb = np.linalg.inv(R_W_G.T) @ R_lb
        R_ub = np.linalg.inv(R_W_G.T) @ R_ub

        self.g_lb = sc.Rotation.from_matrix(R_ub).as_euler("xyz")
        self.g_ub = sc.Rotation.from_matrix(R_lb).as_euler("xyz")



import numpy as np

"""
Finite Differencing Matrix to produce sum of squared accelerations
ddq= A * q
ddq.T * ddq = ddq.T * (A.T * A) * ddq
ddq.T * ddq = ddq.T * P * ddq
"""

class SquaredAccCost():
    def __init__(self, num_waypoints, num_dof, manip_frame = False):
        self._n_waypoints = num_waypoints
        self._n_dof = num_dof 
        self._manip_frame = manip_frame
        self.P = self.create_P_matrix(self._n_waypoints, self._n_dof)

    def create_P_matrix(self, num_waypoints, num_dof):
        if self._manip_frame:
            FD_matrix = np.zeros((num_waypoints, num_waypoints), dtype=float)
            FD_INDICES = [1., -2., 1.]
            for i in range(1, num_waypoints-1):
                FD_matrix[i, (i-1):(i-1)+3] = FD_INDICES
            FD_matrix[0,0]=1
            FD_matrix[-1,-1] = 1
            R = FD_matrix.T@FD_matrix
            I = np.eye(num_dof)
            I[-6:] = 0.0 # dont want to penalize q_dd for grasp frame
            return np.kron(R, I)
        else:
            FD_matrix = np.zeros((num_waypoints, num_waypoints), dtype=float)
            FD_INDICES = [1., -2., 1.]
            for i in range(1, num_waypoints-1):
                FD_matrix[i, (i-1):(i-1)+3] = FD_INDICES
            FD_matrix[0,0]=1
            FD_matrix[-1,-1] = 1
            R = FD_matrix.T @ FD_matrix
            return np.kron(R, np.eye(num_dof))

    def eval_cost(self, q):
        return q.T @ self.P @ q


import numpy as np
import casadi as ca
import spatial_casadi as sc
from grasp_planning.constraints.constraint_template import Constraint

class GraspPosConstraint(Constraint):
    def __init__(self, robot, q_ca, waypoint_ID, n_dofs, tolerance=0.0) -> None:
        super().__init__() 
        # q = q_ca[:n_dof, waypoint_ID] -> joint space of the robot (decision variable)
        # manip_frame = q_ca[n_dof:, waypoint_ID] -> manipulation frame (decision variable)
        # manip frame_pos - T_W_Grasp_pos = 0
        self._robot = robot
        self.g = q_ca[n_dofs:(n_dofs+3), waypoint_ID]
        self.g_eval = ca.Function('g_grasp_pos', [q_ca], [self.g])
        self.tolerance = tolerance
        # self.update_limits(T_W_Grasp)
      
    def get_constraint(self):
        return self.g

    def do_eval(self, q):
        return self.g_eval(q)
    
    def set_limits(self, lb, ub):
        self.g_lb = lb
        self.g_ub = ub

    def get_limits(self):
        return self.g_lb, self.g_ub
    
    def update_limits(self, T_W_Grasp):
        R_W_Grasp_rpy = sc.Rotation.from_matrix(T_W_Grasp[:3,:3]).as_euler("xyz") #convert rotation part of FK into rpy 
        self._T_W_Grasp_rpy = np.vstack((T_W_Grasp[:3,3][:,None], R_W_Grasp_rpy))
        g_lb = (np.zeros((3,1)) + self._T_W_Grasp_rpy[:3]) - self.tolerance
        g_ub = (np.zeros((3,1)) + self._T_W_Grasp_rpy[:3]) + self.tolerance
        self.set_limits(g_lb, g_ub)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['grasp_planning',
 'grasp_planning.constraints',
 'grasp_planning.cost',
 'grasp_planning.solver']

package_data = \
{'': ['*']}

install_requires = \
['casadi>3.5',
 'forwardkinematics>=1.1.3',
 'numpy>=1.15.3,<2.0.0',
 'scipy<=1.10.1',
 'spatial-casadi>=1.1.0,<2.0.0']

setup_kwargs = {
    'name': 'grasp_planning',
    'version': '0.4.1',
    'description': '"Grasp and Motion Planning Python Package."',
    'author': 'Tomas Merva',
    'author_email': 'tmerva7@gmail.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)

